// Run one test with -g + keyword:
//mocha -g syntax

//-----------------------------------------------------------------------------
// Init
//-----------------------------------------------------------------------------

const should  = require('should');
const sinon   = require('sinon');
const sandbox = require('./sandbox.js')


//-----------------------------------------------------------------------------
// Tests
//-----------------------------------------------------------------------------

describe('Sandbox', function(){
  // beforeEach(function(){
  //   sandbox = new sandbox()
  // });

  it('Loads', function(){
    // import './sandbox.js'
    sandbox('', null, null)
    var a = '2'
    a.should.eql('2');

  });

  it('should run simple addition', function(done) {
    sandbox('RESULT = 1+1', null, null, function(err, res ){
      console.log('res', res)
      res.result.should.eql(2);
       done();
    })
  });

  it('should run simple assignment', function(done) {
    sandbox('a=1;RESULT = 1+a', null, null, function(err, res ){
      console.log('res', res)
      res.result.should.eql(2);
       done();
    })
  });

  it('should run multiline text', function(done) {
    sandbox('a=1 \n RESULT = 1+a', null,  null,function(err, res ){
      console.log('res', res)
      res.result.should.eql(2);
       done();
    })
  });

  it('should recognize objects', function(done) {
    sandbox('obj = {a:1} \n RESULT = 1+obj.a', null,  null, function(err, res ){
      console.log('res', res)
      res.result.should.eql(2);
       done();
    })
  });

  it('handles mysterious issue with literals', function(done) {
    var myCode = 'RESULT = "f"'

    sandbox( myCode , null, null, function(err, res ){

      console.log('res', res)
      console.log('err', err)

      res.result.should.eql("f");
      done();
    })
  });

  it('should gracefully handle syntax errors', function(done) {
    sandbox('hi )there', null,  null,function(err, res ){
      console.log('err', err)
      err.result.should.eql('SyntaxError : Unexpected token )');
       done();
    })
  });

  it('should effectively prevent code from accessing node.js ', function(done) {
    sandbox('RESULT = process.platform', null,  null,function(err, res ){
      err.result.should.eql('ReferenceError : process is not defined');
       done();
    })
  });

  it('should effectively prevent code from circumventing the sandbox', function(done) {
    sandbox("var sys=require('sys'); sys.puts('Up in your fridge')", null,  null,function(err, res ){
      // console.log('res', res)
      err.result.should.eql('ReferenceError : require is not defined');
       done();
    })

  });

  it('sends console logs to sandbox parent on success', function(done) {
    var myCode = 'console.log("hello" ); RESULT=1'

    sandbox( myCode , null, null, function(err, res ){

      console.log('res', res)
      console.log('err', err)

      res.console[0].should.eql("hello\n");
      done();
    })
  });

  it('sends console logs to sandbox parent, on failure', function(done) {
    var myCode = 'console.log("hello" ); failNow();  RESULT=1'

    sandbox( myCode , null, null, function(err, res ){

      console.log('res', res)
      console.log('err', err)

      err.console[0].should.eql("hello\n");
      done();
    })
  });

  it('can load whitelisted libraries', function(done) {
    var myCode = 'fs.readFileSync("package.json") ; RESULT = 42'
    var myLibraries = [{ namespace: 'fs', npm: 'fs'}]

    sandbox( myCode , myLibraries, null, function(err, res ){
      console.log('res', res)
      res.result.should.eql(42);
       done();
    })
  });

  it('can load MULTIPLE white-listed libraries in the default context', function(done) {
    var myCode = 'fs.readFileSync("package.json") ; RESULT = util.isArray([])'
    var myLibraries = [{ namespace: 'fs', npm: 'fs'}, { namespace: 'util', npm: 'util'} ]

    sandbox( myCode , myLibraries, null, function(err, res ){
      console.log(err);
      console.log('res', res)
      res.result.should.eql(true);
       done();
    })
  });

  it('will not execute a dependency NOT in the default context', function(done) {
    // sb.run(' util.isNumber(false) ', function(output) {
    //   output.result.should.eql('\'ReferenceError: util is not defined\'');
    //   done();
    // });
    var myCode = 'fs.readFileSync("package.json") '
    var myLibraries = [{ namespace: 'util', npm: 'util'} ] //fs not here

    sandbox( myCode , myLibraries, null, function(err, res ){
      console.log('err', err)
      err.result.should.eql('ReferenceError : fs is not defined');
       done();
    })

  });

  it('can append an object context to the defaultContext ', function(done) {
    sandbox( 'RESULT = test' , null, {test: 1234} , function(err, res ){
      // console.log('err', err)
      res.result.should.eql(1234);
       done();
    })
  });

  it('can append a function to defaultContext ', function(done) {
    sandbox( 'RESULT = test()' , null, {test: function(){ return 1234} } , function(err, res ){
      console.log('err', err)
      console.log('res', res)

      res.result.should.eql(1234);
       done();
    })
  });

  // can wrap functions, so that in wireline the sandbox runner can wrap the functions so that only the RESULT variable is exposed in the sandbox
  it('can wrap functions', function(done) {

    var script= 'function wrapped(){\
\n\tfunction add(first, second){\
\n\treturn first + second;\
\n}\
\nreturn add(20, 22)\
\n}\
\n RESULT = wrapped()'

    sandbox( script , null, {} , function(err, res ){
      console.log('err', err)
      console.log('res', res)

      res.result.should.eql(42);
       done();
    })
  });


  it('can have wrapped functions access INPUT object', function(done) {

    var script= 'function wrapped(){\
  \n\tfunction add(first, second){\
  \n\treturn first + Number(INPUT.a);\
  \n}\
  \nreturn add(20, 22)\
  \n}\
  \n RESULT = wrapped()'

    sandbox( script , null, { INPUT: { a: 22} }  , function(err, res ){
      console.log('err', err)
      console.log('res', res)

      res.result.should.eql(42);
       done();
    })
  });

  // it('TODO: can have wrapped function using "done()" ', function(done) {
  //   // var myCode = "x=1;rp('http://www.google.com').then( function(){x=2 \n callback() })"
  //   var myCode = "function botFunction(){\
  //     a=1; setTimeout(function(){\
  //       a=2;\
  //       return a\
  //     }, 10);\
  //     return a\
  //   };\
  //     RESULT = botFunction() "
  //
  //   myLibraries = []
  //   sandbox( myCode , myLibraries, null, function(err, res ){
  //     console.log('res: ', res);
  //     res.result.should.eql(2);
  //     done();
  //   })
  // });

  it('done() async api works with long-running async functions', function(done) {
    var myCode = "RESULT=1;setTimeout(function(){ RESULT=2 \n done()}, 100)"
    // var myCode = "a=3 \n RESULT = 1+a \n done()"

    myLibraries = [{ namespace: 'rp', npm: 'request-promise'}]
    sandbox( myCode , myLibraries, null, function(err, res ){
      res.result.should.eql(2);
      done();
    })
  });

  it('done() can return a value', function(done) {
    var myCode = "RESULT=1; setTimeout(function(){ done( 2 ) }, 100)"
    // var myCode = "a=3 \n RESULT = 1+a \n done()"

    myLibraries = [{ namespace: 'rp', npm: 'request-promise'}]
    sandbox( myCode , myLibraries, null, function(err, res ){
      res.result.should.eql(2);
      done();
    })
  });

  it('should run async libraries like request-promise', function(done) {
    var myCode = "rp('http://mockbin.org/bin/17889536-56e3-498d-9caa-be279a7e5a8d/').then( function (htmlString) {console.log(htmlString); RESULT = JSON.parse(htmlString) ; done() })"
    myLibraries = [{ namespace: 'rp', npm: 'request-promise'}]

    sandbox( myCode , myLibraries, null, function(err, res ){

      console.log('res', res)
      res.result.Hello.should.eql('World');
      done();
    })
  });

  it('uses in-house Request promise', function(done) {
    var myCode = "httpRequest({ url: 'http://mockbin.org/bin/17889536-56e3-498d-9caa-be279a7e5a8d/' }).then(function( body ){ RESULT = body; done()})"
    // myLibraries = [{ namespace: 'rp', npm: 'request-promise'}]

    sandbox( myCode , null, null, function(err, res ){

      console.log('res', res.result)
      result = JSON.parse( res.result )
      console.log('parsed res', result);
      result.Hello.should.eql('World');
      done();
    })
  });


  //TODO:
  it('Unknown errors reported ', function(done) {
    var myCode = "function test(){ return INPUT[0] }; RESULT = test()"

    sandbox( myCode , null, { INPUT: { a: 22} }, function(err, res ){

      console.log('res', res)
      console.log('err', err)

      // result = JSON.parse( res.result )
      // console.log('parsed res', result);
      res.status.should.eql(  'COMPLETE' );
      done();
    })
  });


  it('Allows variable assignment within Request', function(done) {
    var myCode = "var assignment = 'useless'; httpRequest({ url: 'http://mockbin.org/bin/17889536-56e3-498d-9caa-be279a7e5a8d/' }).then(function( body ){ RESULT = body; done()})"

    sandbox( myCode , null, null, function(err, res ){

      console.log('res', res)
      result = JSON.parse( res.result )
      console.log('parsed res', result);
      result.Hello.should.eql('World');
      done();
    })
  });


    it('returns undefined library correctly', function(done) {

      var script= 'function wrapped(){\
    \nreturn request({ url: "http://mockbin.org/bin/17889536-56e3-498d-9caa-be279a7e5a8d/" }).then(function( body ){ RESULT = body; done()})\
    \n};\
    \n wrapped()'

      sandbox( script , null, { INPUT: { a: 22} }  , function(err, res ){
        console.log('sandbox2test.js - err', err)
        console.log('sandbox2test.js - res', res)
        err.status.should.eql('ERROR');
        done();
      })
    });

    // it('TODO: issue identifying async done vs return', function(done) {
    //   var myCode = "RESULT=1;\
    //   if ( false ){\
    //     setTimeout(function(){\
    //        RESULT=2; \
    //        done()\
    //      }, 100)\
    //   }else{\
    //     RESULT=2;\
    //   }"
    //   // var myCode = "a=3 \n RESULT = 1+a \n done()"
    //
    //   myLibraries = [{ namespace: 'rp', npm: 'request-promise'}]
    //   sandbox( myCode , myLibraries, null, function(err, res ){
    //     res.result.should.eql(2);
    //     done();
    //   })
    // });

    it('TODO: issue identifying async done vs return', function(done) {
      var myCode= 'function wrapped(){\
  \n\tfunction add(first, second){\
  \n\treturn first + second;\
  \n}\
  \nreturn add(20, 22)\
  \n}\
  \n RESULT = wrapped()'
      // var myCode = "a=3 \n RESULT = 1+a \n done()"

      myLibraries = [{ namespace: 'rp', npm: 'request-promise'}]
      sandbox( myCode , myLibraries, null, function(err, res ){
        res.result.should.eql(2);
        done();
      })
    });

    // it('TODO: process_env', function(done) {
    //   var myCode = "console.log('process inside'); RESULT = 2;"
    //   // var myCode = "a=3 \n RESULT = 1+a \n done()"
    //
    //   myLibraries = [{ namespace: 'rp', npm: 'request-promise'}]
    //   sandbox( myCode , myLibraries, null, function(err, res ){
    //     res.result.should.eql(2);
    //     done();
    //   })
    // });
    //
    // it('TODO: testing', function(done) {
    //   var myCode = "\
    //      done( (function(){ \
    //        RESULT=1;\
    //        if ( true ){\
    //          setTimeout(function(){\
    //             RESULT=2; \
    //             done()\
    //           }, 100)\
    //        }else{\
    //          RESULT=2;\
    //          return\
    //        }\
    //      } )() )\
    //   "

      // var myCode = "RESULT=1;\
      // if ( true ){\
      //   setTimeout(function(){\
      //      RESULT=2; \
      //      done()\
      //    }, 100)\
      // }else{\
      //   RESULT=2;\
      // }"
      // var myCode = "a=3 \n RESULT = 1+a \n done()"

      // myLibraries = [{ namespace: 'rp', npm: 'request-promise'}]
      // sandbox( myCode , myLibraries, null, function(err, res ){
      //   res.result.should.eql(2);
      //   done();
      // })
    // });
});
